;;Common Lisp Interactive Fiction Library 
;;
;;verb-lib module: defines verbs and their associated actions
;;
;;This file is a part of Lisp Interactive Fiction Project
;;
;;See license.txt for licensing information



(in-package :cl-user)

(defpackage :verb-lib
  (:use :common-lisp :if-lib :if-basic-lib)
  (:export :attack :take :teleport :examine 
	   :go-to 
	   :take :put-in :put-on :drop :receive
	   :wear :strip :enter :climb)
  (:shadow :listen)
  (:shadowing-import-from :if-lib :room))

(in-package :verb-lib)

(defmacro const-fun (name args value)
  `(defun ,name ,args
    (declare (ignore ,@args))
    ,value))

(const-fun noargs-1 (c) nil)

(verb "quit" '(:meta -> quit-game noargs-1)) ;;That one you'll use often ;)

;Debug verb
(verb "teleport"
      `((:noun ,(lambda () *allobjects*)) -> teleport))

(verb "take"
      '(:noun -> take)
      '("off" :held -> strip)
      '(:held "off" -> strip)) 

(verb "get"
      '(:noun -> take))
   
(const-fun const-loc (c) *location*)

(verb "look"
      `(-> look const-loc)
      '("at" :seen -> examine))

(verb "examine" "x"
      '(:noun -> examine))

(verb "attack" "break" "crack" "destroy"
     "fight" "hit" "kill" "murder" "punch"
     "smash" "thump" "torture" "wreck"
    '(:noun -> attack))

;(defmacro const-fun* (name args value)
;  `(defun ,name ,args
;    (declare (ignore ,@args))
;    (list *location* ,value)))

(const-fun cdir-n (c) dir-n)
(const-fun cdir-ne (c) dir-ne)
(const-fun cdir-e (c) dir-e)
(const-fun cdir-se (c) dir-se)
(const-fun cdir-s (c) dir-s)
(const-fun cdir-sw (c) dir-sw)
(const-fun cdir-w (c) dir-w)
(const-fun cdir-nw (c) dir-nw)
(const-fun cdir-u (c) dir-u)
(const-fun cdir-d (c) dir-d)
(const-fun cdir-in (c) dir-in)
(const-fun cdir-out (c) dir-out)

(verb "go" "run" "walk" 
      '(:direction -> go-to)
      '(:noun -> enter)
      '((:or "into" "in" "inside" "through") :noun -> enter rest))

(verb "n" "north" '(-> go-to cdir-n))
(verb "ne" "northeast" '(-> go-to cdir-ne))
(verb "e" "east" '(-> go-to cdir-e))
(verb "se" "southeast" '(-> go-to cdir-se))
(verb "s" "south" '(-> go-to cdir-s))
(verb "sw" "southwest" '(-> go-to cdir-sw))
(verb "w" "west" '(-> go-to cdir-w))
(verb "nw" "northwest" '(-> go-to cdir-nw))
(verb "u" "up" '(-> go-to cdir-u))
(verb "d" "down" '(-> go-to cdir-d))
(verb "in" '(-> go-to cdir-in))
(verb "out" '(-> go-to cdir-out))

(verb "enter" 
      '(:direction -> go-to)
      '(:noun -> enter))

(verb "inventory" "i" '(-> inventory))

(verb "take"
      '(:noun -> take)
      '("off" :held -> strip)
      '(:held "off" -> strip))

(verb "get"
      '(:noun -> take)
      '((:or "in" "into" "on" "onto") :noun -> enter rest))

(verb "drop" "discard" "throw"
      '(:held -> drop)
      '(:held "in" :noun -> put-in)
      '(:held "on" :noun -> put-on))
      
(verb "put"
      '(:held "on" :noun -> put-on)
      '(:held "in" :noun -> put-in)
      '(:held "down" -> drop)
      '("on" :held -> wear)
      '(:held -> drop))

(verb "wear" "don"
      '(:held -> wear))

(verb "remove"
      '(:held -> strip)
      '(:noun -> take))

(verb "shed" "disrobe" "doff"
      '(:held -> strip))

(verb "sit" "lie"
      '("on" "top" "of" :noun -> enter)
      '((:or "on" "in" "inside") :noun -> enter rest))

(verb "climb" "scale"
      '(:noun -> climb)
      '((:or "up" "over") :noun -> climb))

(verb "listen" "hear"
      '(-> listen const-loc)
      '(:noun -> listen)
      '("to" :noun -> listen))

(defaction attack (obj) "Violence is not the answer.")

(defaction teleport (obj) 
  (go-to-room obj))

(defaction examine (obj)
  (if (provides obj 'description)
      (read-property obj 'description)
      (format nil "You see nothing special about ~A.~%" (the-name obj))))  

;;(defun look-around () (run-action 'look *location*))

(defaction go-to (dir)
  (let ((destination (read-property *location* (property dir))))
    (if destination (go-to-room destination)
	(if (provides *location* 'cant-go) 
	    (read-property *location* 'cant-go)
	    "You can't go here."))))

;; (defaction go-n () (run-action 'go-to dir-n))
;; (defaction go-ne () (run-action 'go-to dir-ne))
;; (defaction go-e () (run-action 'go-to dir-e))
;; (defaction go-se () (run-action 'go-to dir-se))
;; (defaction go-s () (run-action 'go-to dir-s))
;; (defaction go-sw () (run-action 'go-to dir-sw))
;; (defaction go-w () (run-action 'go-to dir-w))
;; (defaction go-nw () (run-action 'go-to dir-nw))
;; (defaction go-u () (run-action 'go-to dir-u))
;; (defaction go-d () (run-action 'go-to dir-d))
;; (defaction go-in () (run-action 'go-to dir-in))
;; (defaction go-out () (run-action 'go-to dir-out))

(defun inventory ()
  (sprint "You are carrying: ~a." (list-contents *player*))
  (newline))

(defaction take (obj)
  "You can't take that.")

(defmethod take((obj item))
  (if (has obj :item)
      (if (in obj *player*) 
	  (progn (sprint "You already have ~A" (the-name obj)) t) 
	  (progn 
	    (move obj *player*)
	    (when (run-action-after obj) "Taken.")))
      (call-next-method)))

(defaction drop (obj)
  (unless (has obj :item) (return-from drop "You can't drop that."))
  (when (has obj :worn)
    (sprint "(first removing ~a)~%" (the-name obj))
    (unless (run-action 'strip obj)
      (return-from drop "You can't drop it.")))  
  (move obj (parent *player*))
  (when (run-action-after obj) "Dropped."))

(defaction put-on (item host)
  "You can't put anything on that.")

(defmethod put-on ((item item) (host supporter))
  ;;(format t "(~a ~a)" (print-name item) (print-name host)) 
  (unless (has item :item) (return-from put-on "You can't get rid of that."))
  (unless (has host :supporter) (return-from put-on (call-next-method)))
  (and (run-action 'receive (reverse *args*) :time 0)
       *after*
       (run-action-after item) 
       "Done."))

(defaction put-in (item host)
  "You can't put anything in that.")

(defmethod put-in ((item item) (host container))
  (unless (has item :item) (return-from put-in "You can't get rid of that."))
  (unless (has host :container) (return-from put-in (call-next-method)))
  (when (has host :closed) 
    (return-from put-in 
      (format nil "~a is closed." (the-name host :capital t))))
  (and (run-action 'receive (reverse *args*) :time 0)
       *after*
       (run-action-after item) 
       "Done."))
    
(defaction receive (host guest)
  "No method defined for that kind of object movement.")

(defmethod receive ((host supporter) (item item))
  (if (or (zerop (capacity host)) 
	  (< (list-length (children host)) (capacity host)))
    (progn (move item host)
	   (run-action-after host))
    "Not enough space."))

(defmethod receive ((host container) (item item))
  (if (or (zerop (capacity host)) 
	  (< (list-length (children host)) (capacity host)))
    (progn (move item host)
	   (run-action-after host))
    "Not enough space."))

(defaction wear (what)
  "You can't wear that.")

(defmethod wear ((obj clothing))
  (if (has obj :clothing)
      (if (hasnt obj :worn) 
	  (progn 
	    (give obj :worn) (when (run-action-after obj) "Done."))
	  "You are already wearing it.")
      "You can't wear that."))

(defaction strip (what)
  "That's one strange thing you want to do.")

(defmethod strip ((obj clothing))
  (if (and (has obj :clothing) (has obj :worn))
      (progn (give obj :~worn) (when (run-action-after obj) "Done."))
      "You can't do that."))

(defaction enter (what)
  "You can't enter that.")

(defaction climb (what)
  "You can't climb that.")

(defaction listen (what)
  "You hear nothing unexpected.")